#include "xrtcpsu.h"

XRtcPsu_Config XRtcPsu_ConfigTable[] __attribute__ ((section (".drvcfg_sec"))) = {

	{
		"xlnx,zynqmp-rtc", /* compatible */
		0xf12a0000, /* reg */
		{0x40c8,  0x40c9}, /* interrupts */
		0xe2000000 /* interrupt-parent */
	},
	 {
		 NULL
	}
};